@@ -2,7 +2,7 @@
  * way so that they may be used in slang scripts.
  */
 /*
-Copyright (C) 2004, 2005, 2006, 2007, 2008 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -41,7 +41,7 @@ static Errno_Map_Type Errno_Map [] =
 #ifndef EPERM
 # define EPERM	-1
 #endif
-     {"Not owner",			EPERM,	"EPERM"},
+     {"Operation not permitted",	EPERM,	"EPERM"},
 #ifndef ENOENT
 # define ENOENT	-1
 #endif
