@@ -1,6 +1,6 @@
 /* List of objects */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -42,22 +42,25 @@ USA.
 
 typedef struct _pSLang_List_Type SLang_List_Type;
 
-#define CHUNK_SIZE 32
+#define CHUNK_SIZE 128
 
 typedef struct _Chunk_Type
 {
    struct _Chunk_Type *next;
    struct _Chunk_Type *prev;
    int num_elements;
-   SLang_Object_Type elements[CHUNK_SIZE];
+   SLang_Object_Type *elements;	       /* CHUNK_SIZE of em */
 }
 Chunk_Type;
 
 struct _pSLang_List_Type
 {
-   int length;
+   SLindex_Type length;
    Chunk_Type *first;
    Chunk_Type *last;
+   
+   Chunk_Type *recent;		       /* most recent chunk accessed */
+   SLindex_Type recent_num;	       /* num elements before the recent chunk */
 };
 
 static void delete_chunk (Chunk_Type *c)
@@ -71,8 +74,8 @@ static void delete_chunk (Chunk_Type *c)
    n = c->num_elements;
    objs = c->elements;
    for (i = 0; i < n; i++)
-     SLang_free_object (&objs[i]);
-   
+     SLang_free_object (objs+i);
+   SLfree ((char *) objs);
    SLfree ((char *) c);
 }
 
@@ -81,6 +84,15 @@ static Chunk_Type *new_chunk (void)
    Chunk_Type *c;
    
    c = (Chunk_Type *) SLcalloc (1, sizeof (Chunk_Type));
+   if (c == NULL)
+     return c;
+
+   c->elements = (SLang_Object_Type *)SLcalloc(CHUNK_SIZE, sizeof(SLang_Object_Type));
+   if (c->elements == NULL)
+     {
+	SLfree ((char *) c);
+	return NULL;
+     }
    return c;
 }
 
@@ -145,9 +157,10 @@ static SLang_List_Type *allocate_list (void)
 
 static SLang_Object_Type *find_nth_element (SLang_List_Type *list, int nth, Chunk_Type **cp)
 {
-   Chunk_Type *c;
+   Chunk_Type *c, *first, *last;
    int n, next_n;
    int length;
+   int dir;
 
    length = list->length;
    if (nth < 0)
@@ -155,13 +168,56 @@ static SLang_Object_Type *find_nth_element (SLang_List_Type *list, int nth, Chun
    
    if ((nth < 0) || (nth >= length))
      {
-	SLang_verror (SL_Index_Error, "List Index out of range");
+	_pSLang_verror (SL_Index_Error, "List Index out of range");
 	return NULL;
      }
-   if (nth < length/2)
+
+   n = 0;
+   first = list->first;
+   last = list->last;   
+   dir = 1;
+
+   if (list->recent != NULL)
      {
-	n = 0;
-	c = list->first;
+	int mid = list->recent_num;
+	if (nth < mid)
+	  {
+	     if (nth > mid/2)
+	       {
+		  dir = -1;
+		  n = mid;
+		  last = list->recent->prev;
+		  /* Note: last cannot be NULL.  The only way for it to be NULL 
+		   * is for list->recent == list_first, which would mean that
+		   * mid==0.  But the condition here is that nth < mid, and
+		   * nth >= 0
+		   */
+	       }
+	  }
+	else
+	  {
+	     if (nth > (mid/2 + length/2))
+	       {
+		  dir = -1;
+		  n = length;
+	       }
+	     else
+	       {
+		  dir = 1;
+		  n = mid;
+		  first = list->recent;
+	       }
+	  }
+     }
+   else if (nth > length/2)
+     {
+	n = length;
+	dir = -1;
+     }
+
+   if (dir > 0)
+     {
+	c = first;
 	while (nth >= (next_n = n + c->num_elements))
 	  {
 	     n = next_n;
@@ -169,11 +225,14 @@ static SLang_Object_Type *find_nth_element (SLang_List_Type *list, int nth, Chun
 	  }
 	if (cp != NULL)
 	  *cp = c;
+
+	list->recent = c;
+	list->recent_num = n;
+
 	return c->elements + (nth - n);
      }
 
-   n = length;
-   c = list->last;
+   c = last;
    while ((next_n = n - c->num_elements) > nth)
      {
 	n = next_n;
@@ -182,6 +241,9 @@ static SLang_Object_Type *find_nth_element (SLang_List_Type *list, int nth, Chun
    
    if (cp != NULL)
      *cp = c;
+   
+   list->recent = c;
+   list->recent_num = next_n;
    return c->elements + (nth - next_n);
 }
 
@@ -196,87 +258,6 @@ static int pop_list (SLang_MMT_Type **mmtp, SLang_List_Type **list)
    return 0;
 }
 
-static int pop_list_and_index (unsigned int num_indices, 
-			       SLang_MMT_Type **mmtp, SLang_List_Type **listp,
-			       int *indx)
-{
-   int idx;
-   SLang_MMT_Type *mmt;
-   SLang_List_Type *list;
-
-   if (-1 == pop_list (&mmt, &list))
-     return -1;
-
-   if (num_indices != 1)
-     {
-	SLang_verror (SL_InvalidParm_Error, "List_Type objects are limited to a single index");
-	SLang_free_mmt (mmt);
-	return -1;
-     }
-
-   if (-1 == SLang_pop_integer (&idx))
-     {
-	SLang_free_mmt (mmt);
-	return -1;
-     }
-   *indx = idx;
-   *listp = list;
-   *mmtp = mmt;
-   return 0;
-}
-
-/* FIXME: Extend this to allow an index array */
-static int _pSLlist_aget (SLtype type, unsigned int num_indices)
-{
-   SLang_MMT_Type *mmt;
-   SLang_List_Type *list;
-   SLang_Object_Type *obj;
-   int ret = 0;
-   int indx;
-
-   (void) type;
-
-   if (-1 == pop_list_and_index (num_indices, &mmt, &list, &indx))
-     return -1;
-   
-   obj = find_nth_element (list, indx, NULL);
-   if (obj != NULL)
-     ret = _pSLpush_slang_obj (obj);
-
-   SLang_free_mmt (mmt);
-   return ret;
-}
-
-static int _pSLlist_aput (SLtype type, unsigned int num_indices)
-{
-   SLang_MMT_Type *mmt;
-   SLang_List_Type *list;
-   SLang_Object_Type obj;
-   SLang_Object_Type *elem;
-   int indx;
-
-   (void) type;
-
-   if (-1 == pop_list_and_index (num_indices, &mmt, &list, &indx))
-     return -1;
-
-   if (-1 == SLang_pop (&obj))
-     {
-	SLang_free_mmt (mmt);
-	return -1;
-     }
-   
-   if (NULL == (elem = find_nth_element (list, indx, NULL)))
-     {
-	SLang_free_object (&obj);
-	SLang_free_mmt (mmt);
-	return -1;
-     }
-   SLang_free_object (elem);
-   *elem = obj;
-   SLang_free_mmt (mmt);
-   return 0;
-}
 
 static SLang_List_Type *make_sublist (SLang_List_Type *list, int indx_a, int indx_b)
 {
@@ -300,7 +281,7 @@ static SLang_List_Type *make_sublist (SLang_List_Type *list, int indx_a, int ind
    
    if ((indx_b >= length) || (indx_a < 0))
      {
-	SLang_verror (SL_Index_Error, "Indices are out of range for list object");
+	_pSLang_verror (SL_Index_Error, "Indices are out of range for list object");
 	return NULL;
      }
 
@@ -358,15 +339,19 @@ static SLang_List_Type *make_sublist (SLang_List_Type *list, int indx_a, int ind
    return new_list;
 }
 
-static void list_delete_elem (SLang_List_Type *list, int *indx)
+static void list_delete_elem (SLang_List_Type *list, int *indxp)
 {
    SLang_Object_Type *elem;
    Chunk_Type *c;
    char *src, *dest, *src_max;
+   int indx;
 
-   if (NULL == (elem = find_nth_element (list, *indx, &c)))
+   indx = *indxp;
+   if (NULL == (elem = find_nth_element (list, indx, &c)))
      return;
 
+   if (indx < 0) indx += list->length; /* checked by find_nth_element */
+
    SLang_free_object (elem);
    c->num_elements--;
    list->length--;
@@ -382,6 +367,8 @@ static void list_delete_elem (SLang_List_Type *list, int *indx)
 	if (c->prev != NULL)
 	  c->prev->next = c->next;
 	delete_chunk (c);
+	if (list->recent == c)
+	  list->recent = NULL;
 	return;
      }
 
@@ -390,11 +377,9 @@ static void list_delete_elem (SLang_List_Type *list, int *indx)
    src_max = src + sizeof(SLang_Object_Type)*((c->elements+c->num_elements)-elem);
    while (src < src_max)
      *dest++ = *src++;
-
-#if 0
-   memcpy ((char *)elem, (char *)(elem+1), 
-	   sizeof(SLang_Object_Type)*((c->elements+c->num_elements)-elem));
-#endif
+   
+   if ((list->recent != NULL) && (list->recent_num > indx))
+     list->recent_num--;
 }
 
 static void slide_right (Chunk_Type *c, int n)
@@ -411,11 +396,12 @@ static void slide_right (Chunk_Type *c, int n)
      }
 }
 
-static int insert_nonlist_element (SLang_List_Type *list, SLang_Object_Type *obj, int indx)
+static int insert_element (SLang_List_Type *list, SLang_Object_Type *obj, int indx)
 {
    Chunk_Type *c, *c1;
    SLang_Object_Type *elem;
    int num;
+   int equality_ok = 0;
 
    if (indx == 0)
      {
@@ -437,6 +423,8 @@ static int insert_nonlist_element (SLang_List_Type *list, SLang_Object_Type *obj
 	if (list->last == NULL)
 	  list->last = c;
 	c->elements[0] = *obj;
+
+        equality_ok = 1;
 	goto the_return;
      }
 
@@ -456,7 +444,7 @@ static int insert_nonlist_element (SLang_List_Type *list, SLang_Object_Type *obj
 	c->elements[0] = *obj;
 	goto the_return;
      }
-   
+
    if (NULL == (elem = find_nth_element (list, indx, &c)))
      return -1;
 
@@ -482,6 +470,7 @@ static int insert_nonlist_element (SLang_List_Type *list, SLang_Object_Type *obj
 	  list->first = c1;
 	c = c1;
 	c->elements[0] = *obj;
+	equality_ok = 1;
 	goto the_return;
      }
    c1->prev = c;
@@ -492,13 +481,21 @@ static int insert_nonlist_element (SLang_List_Type *list, SLang_Object_Type *obj
    if (list->last == c)
      list->last = c1;
 
-   memcpy ((char *)c1->elements, (char *)elem, sizeof(SLang_Object_Type)*num);
+   memcpy ((char *)c1->elements, (char *)elem, num*sizeof(SLang_Object_Type));
    c1->num_elements = num;
    c->num_elements -= num;
    c->elements[c->num_elements] = *obj;
    /* drop */
 
-   the_return:
+the_return:
+
+   if (list->recent != NULL)
+     {
+	if ((list->recent_num > indx) 
+	    || ((list->recent_num == indx) && equality_ok))
+	  list->recent_num++;
+     }
+
    c->num_elements++;
    list->length++;
    return 0;
@@ -507,77 +504,6 @@ static int insert_nonlist_element (SLang_List_Type *list, SLang_Object_Type *obj
 /* Upon sucess, obj is stored in list and calling routine should not free it.
  * Upon failure, calling routine should free obj.
  */
-#define HAS_LISTS_OF_LISTS 1
-#if HAS_LISTS_OF_LISTS
-#define insert_element insert_nonlist_element
-#else
-static int insert_element (SLang_List_Type *list, SLang_Object_Type *obj, int indx)
-{
-   SLang_List_Type *new_list, *tmp_list;
-   SLang_MMT_Type *mmt;
-   Chunk_Type *c;
-
-   if (obj->data_type != SLANG_LIST_TYPE)
-     return insert_nonlist_element (list, obj, indx);
-   
-   mmt = obj->v.ptr_val;
-   if (NULL == (new_list = (SLang_List_Type *)SLang_object_from_mmt (mmt)))
-     return -1;
-
-   if (new_list->length == 0)
-     {
-	SLang_free_mmt (mmt);
-	return 0;
-     }
-
-   if (NULL == (tmp_list = make_sublist (new_list, 0, -1)))
-     return -1;
-
-   c = tmp_list->first;
-   
-   while (c != NULL)
-     {
-	SLang_Object_Type *emin, *emax;
-	SLang_Object_Type new_obj;
-	Chunk_Type *cnext;
-
-	emin = c->elements;
-	emax = emin + c->num_elements;
-	
-	while (emin < emax)
-	  {
-	     if ((-1 == _pSLpush_slang_obj (emin))
-		 || (-1 == SLang_pop (&new_obj)))
-	       {
-		  delete_list (tmp_list);
-		  return -1;
-	       }
-
-	     if (-1 == insert_nonlist_element (list, &new_obj, indx))
-	       {
-		  SLang_free_object (&new_obj);
-		  delete_list (tmp_list);
-		  return -1;
-	       }
-	     emin++;
-	     indx++;
-	  }
-	cnext = c->next;
-	tmp_list->first = cnext;
-	if (cnext != NULL)
-	  cnext->prev = NULL;
-	else 
-	  tmp_list->last = NULL;
-
-	delete_chunk (c);
-	c = cnext;
-     }
-   delete_list (tmp_list);
-
-   SLang_free_mmt (mmt);
-   return 0;
-}
-#endif
    
 static int pop_insert_append_args (SLang_MMT_Type **mmtp, SLang_List_Type **listp,
 				   SLang_Object_Type *obj, int *indx)
@@ -694,6 +620,8 @@ static void list_reverse (SLang_List_Type *list)
 	c->prev = c1;
 	c = c1;
      }
+
+   list->recent = NULL;
 }
 
 static int list_pop_nth (SLang_List_Type *list, int indx)
@@ -728,13 +656,10 @@ static void list_pop (void)
    SLang_free_mmt (mmt);
 }
 
-int _pSLlist_inline_list (void)
+static int pop_as_list_internal (unsigned int count)
 {
-   unsigned int count;
    SLang_List_Type *list;
 
-   count = SLang_Num_Function_Args;
-   
    if (NULL == (list = allocate_list ()))
      return -1;
 
@@ -760,6 +685,76 @@ int _pSLlist_inline_list (void)
    return -1;
 }
 
+static void pop_as_list (int *nitems)
+{
+   if (*nitems < 0)
+     {
+	_pSLang_verror (SL_INVALID_PARM, "Expecting a non-negative integer");
+	return;
+     }
+   (void) pop_as_list_internal (*nitems);
+}
+
+int _pSLlist_inline_list (void)
+{
+   return pop_as_list_internal (SLang_Num_Function_Args);
+}
+
+static void push_list_elements (SLang_List_Type *list)
+{
+   int n = list->length;
+   int i;
+   
+   for (i = 0; i < n; i++)
+     {
+	SLang_Object_Type *objp = find_nth_element (list, i, NULL);
+	if ((objp == NULL) || (-1 == _pSLpush_slang_obj (objp)))
+	  return;
+     }
+}
+
+static int l2a_get_type_callback (VOID_STAR vlist, SLuindex_Type i, SLtype *type)
+{
+   SLang_Object_Type *obj;
+   
+   if (NULL == (obj = find_nth_element ((SLang_List_Type *)vlist, i, NULL)))
+     return -1;
+   *type = obj->o_data_type;
+   return 0;
+}
+
+static int l2a_push_callback (VOID_STAR vlist, SLuindex_Type i)
+{
+   SLang_Object_Type *obj;
+   
+   if (NULL == (obj = find_nth_element ((SLang_List_Type *)vlist, i, NULL)))
+     return -1;
+
+   return _pSLpush_slang_obj (obj);
+}
+
+
+static void list_to_array (void)
+{
+   SLang_List_Type *list;
+   SLang_MMT_Type *mmt;
+   SLtype type = 0;
+   
+   if ((SLang_Num_Function_Args == 2)
+       && (-1 == SLang_pop_datatype (&type)))
+     return;
+   
+   if (-1 == pop_list (&mmt, &list))
+     return;
+
+   (void) _pSLarray_convert_to_array ((VOID_STAR) list, 
+				      l2a_get_type_callback,
+				      l2a_push_callback,
+				      list->length, type);
+
+   SLang_free_mmt (mmt);
+}
+
 #define L SLANG_LIST_TYPE
 #define I SLANG_INT_TYPE
 static SLang_Intrin_Fun_Type Intrin_Table [] =
@@ -768,8 +763,11 @@ static SLang_Intrin_Fun_Type Intrin_Table [] =
    MAKE_INTRINSIC_1("list_reverse", list_reverse, SLANG_VOID_TYPE, L),
    MAKE_INTRINSIC_0("list_insert", list_insert_elem, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("list_append", list_append_elem, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("list_to_array", list_to_array, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("list_new", list_new, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("list_pop", list_pop, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_1("__push_list", push_list_elements, SLANG_VOID_TYPE, L),
+   MAKE_INTRINSIC_1("__pop_list", pop_as_list, SLANG_VOID_TYPE, I),
    SLANG_END_INTRIN_FUN_TABLE
 };
 #undef L
@@ -882,7 +880,7 @@ cl_foreach_open (SLtype type, unsigned int num)
    
    if (num != 0)
      {
-	SLang_verror (SL_NOT_IMPLEMENTED,
+	_pSLang_verror (SL_NOT_IMPLEMENTED,
 		      "%s does not support 'foreach using' form",
 		      SLclass_get_datatype_name (type));
 	return NULL;
@@ -927,7 +925,261 @@ static int cl_foreach (SLtype type, SLang_Foreach_Context_Type *c)
    return 1;
 }
 
+static int pop_list_and_index (unsigned int num_indices, 
+			       SLang_MMT_Type **mmtp, SLang_List_Type **listp,
+			       SLang_Array_Type **ind_atp,
+			       SLindex_Type *indx)
+{
+   SLang_MMT_Type *mmt;
+   SLang_List_Type *list;
+
+   if (-1 == pop_list (&mmt, &list))
+     return -1;
+
+   if (num_indices != 1)
+     {
+	_pSLang_verror (SL_InvalidParm_Error, "List_Type objects are limited to a single index");
+	SLang_free_mmt (mmt);
+	return -1;
+     }
+
+   *ind_atp = NULL;
+   if (SLang_peek_at_stack () == SLANG_ARRAY_INDEX_TYPE)
+     {
+	if (-1 == SLang_pop_array_index (indx))
+	  {
+	     SLang_free_mmt (mmt);
+	     return -1;
+	  }
+     }
+   else
+     {
+	if (-1 == _pSLarray_pop_index (list->length, ind_atp, indx))
+	  {
+	     SLang_free_mmt (mmt);
+	     return -1;
+	  }
+     }
+
+   *listp = list;
+   *mmtp = mmt;
+   return 0;
+}
+
+
+/* FIXME: Extend this to allow an index array */
+static int _pSLlist_aget (SLtype type, unsigned int num_indices)
+{
+   SLang_MMT_Type *mmt;
+   SLang_List_Type *list, *new_list;
+   SLang_Object_Type *obj;
+   SLang_Array_Type *ind_at;
+   SLindex_Type indx, *idx_data;
+   SLuindex_Type i, num, list_len;
+   int ret;
+
+   (void) type;
+
+   if (-1 == pop_list_and_index (num_indices, &mmt, &list, &ind_at, &indx))
+     return -1;
    
+   ret = -1;
+   if (ind_at == NULL)
+     {
+	obj = find_nth_element (list, indx, NULL);
+	if (obj != NULL)
+	  ret = _pSLpush_slang_obj (obj);
+
+	SLang_free_mmt (mmt);
+	return ret;
+     }
+
+   if (NULL == (new_list = allocate_list ()))
+     goto free_and_return;
+   
+   list_len = list->length;
+   num = ind_at->num_elements;
+   idx_data = (SLindex_Type *)ind_at->data;
+   for (i = 0; i < num; i++)
+     {
+	SLang_Object_Type *obja;
+	SLang_Object_Type objb;
+
+	indx = idx_data[i];
+	if (NULL == (obja = find_nth_element (list, idx_data[i], NULL)))
+	  goto free_and_return;
+
+	if (-1 == _pSLslang_copy_obj (obja, &objb))
+	  goto free_and_return;
+
+	if (-1 == insert_element (new_list, &objb, i))
+	  {
+	     SLang_free_object (&objb);
+	     goto free_and_return;
+	  }
+     }
+
+   ret = push_list (new_list);	       /* frees upon error */
+   new_list = NULL;
+ 
+free_and_return:
+
+   if (new_list != NULL)
+     delete_list (new_list);
+   SLang_free_mmt (mmt);
+   SLang_free_array (ind_at);
+   return ret;
+}
+
+static int aput_object (SLang_List_Type *list, SLindex_Type indx, SLang_Object_Type *obj)
+{
+   SLang_Object_Type *elem;
+   
+   if (NULL == (elem = find_nth_element (list, indx, NULL)))
+     return -1;
+
+   SLang_free_object (elem);
+   *elem = *obj;
+   return 0;
+}
+
+static int _pSLlist_aput (SLtype type, unsigned int num_indices)
+{
+   SLang_MMT_Type *mmt;
+   SLang_List_Type *list;
+   SLang_Object_Type obj;
+   SLang_Array_Type *ind_at;
+   SLindex_Type indx, *idx_data;
+   SLuindex_Type i, num, list_len;
+   int ret;
+
+   (void) type;
+
+   if (-1 == pop_list_and_index (num_indices, &mmt, &list, &ind_at, &indx))
+     return -1;
+
+   if (ind_at == NULL)
+     {
+	if (-1 == SLang_pop (&obj))
+	  {
+	     SLang_free_mmt (mmt);
+	     return -1;
+	  }
+   
+	if (-1 == aput_object (list, indx, &obj))
+	  {
+	     SLang_free_object (&obj);
+	     SLang_free_mmt (mmt);
+	     return -1;
+	  }
+	SLang_free_mmt (mmt);
+	return 0;
+     }
+   
+   idx_data = (SLindex_Type *)ind_at->data;
+   num = ind_at->num_elements;
+   list_len = list->length;
+   
+   if (-1 == SLang_pop (&obj))
+     {
+	SLang_free_mmt (mmt);
+	SLang_free_array (ind_at);
+	return -1;
+     }
+
+   ret = -1;
+   
+   if (obj.o_data_type == SLANG_ARRAY_TYPE)
+     {
+	SLang_Array_Type *at = obj.v.array_val;
+	if ((at->num_elements != num) || (at->num_dims != 1))
+	  {
+	     SLang_verror (SL_Index_Error, "Inappropriate array for list[indices]=array expression");
+	     goto free_and_return;
+	  }
+	
+	for (i = 0; i < num; i++)
+	  {
+	     SLang_Object_Type objb;
+	     indx = idx_data[i];
+	
+	     if ((-1 == _pSLarray1d_push_elem (at, i))
+		 || (-1 == SLang_pop (&objb)))
+	       goto free_and_return;
+
+	     if (-1 == aput_object (list, indx, &objb))
+	       {
+		  SLang_free_object (&objb);
+		  goto free_and_return;
+	       }
+	  }
+	ret = 0;
+	goto free_and_return;
+     }
+   
+   if (obj.o_data_type == SLANG_LIST_TYPE)
+     {
+	SLang_MMT_Type *mmt2;
+	SLang_List_Type *list2;
+	
+	mmt2 = obj.v.ref;
+	if (NULL == (list2 = (SLang_List_Type *)SLang_object_from_mmt (mmt2)))
+	  goto free_and_return;
+
+	if (list2->length != (SLindex_Type)num)
+	  {
+	     SLang_verror (SL_Index_Error, "Inappropriate list2 size for list[indices]=list2 expression");
+	     goto free_and_return;
+	  }
+
+	for (i = 0; i < num; i++)
+	  {
+	     SLang_Object_Type *obja;
+	     SLang_Object_Type objb;
+	     indx = idx_data[i];
+	
+	     if (NULL == (obja = find_nth_element (list, idx_data[i], NULL)))
+	       goto free_and_return;
+
+	     if (-1 == _pSLslang_copy_obj (obja, &objb))
+	       goto free_and_return;
+
+	     if (-1 == aput_object (list, indx, &objb))
+	       {
+		  SLang_free_object (&objb);
+		  goto free_and_return;
+	       }
+	  }
+	ret = 0;
+	goto free_and_return;
+     }
+   
+   for (i = 0; i < num; i++)
+     {
+	SLang_Object_Type objb;
+	indx = idx_data[i];
+	
+	if (-1 == _pSLslang_copy_obj (&obj, &objb))
+	  goto free_and_return;
+
+	if (-1 == aput_object (list, indx, &objb))
+	  {
+	     SLang_free_object (&objb);
+	     goto free_and_return;
+	  }
+	SLang_free_mmt (mmt);
+     }
+
+   ret = 0;
+   /* drop */
+free_and_return:
+
+   SLang_free_object (&obj);
+   SLang_free_array (ind_at);
+   SLang_free_mmt (mmt);
+   return ret;
+}
+
 int _pSLang_init_sllist (void)
 {
    SLang_Class_Type *cl;
