@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -35,12 +35,10 @@ static void free_token (_pSLang_Token_Type *t)
 
    if (nrefs == 1)
      {
-	if (t->free_sval_flag)
+	if (t->free_val_func != NULL)
 	  {
-	     if (t->type == BSTRING_TOKEN)
-	       SLbstring_free (t->v.b_val);
-	     else
-	       _pSLfree_hashed_string (t->v.s_val, strlen (t->v.s_val), t->hash);
+	     (*t->free_val_func)(t);
+	     t->free_val_func = NULL;
 	     t->v.s_val = NULL;
 	  }
      }
@@ -65,6 +63,8 @@ static _pSLang_Token_Type Next_Token;
 static int Last_Line_Number = -1;
 #endif
 
+static int In_Looping_Context = 0;
+
 static int unget_token (_pSLang_Token_Type *ctok)
 {
    if (_pSLang_Error)
@@ -96,6 +96,137 @@ static int get_token (_pSLang_Token_Type *ctok)
    return _pSLget_token (ctok);
 }
 
+static int check_int_token_overflow (_pSLang_Token_Type *ctok, int sign)
+{
+   long ival, lval;
+   SLtype stype;
+
+   ctok->v.long_val = lval = sign * ctok->v.long_val;
+
+   if (ctok->flags & SLTOKEN_IS_HEX)
+     return 0;
+
+   switch (ctok->type)
+     {
+      case CHAR_TOKEN:
+	stype = SLANG_CHAR_TYPE;
+	ival = (long)(char) lval;
+	break;
+      case SHORT_TOKEN:
+	stype = SLANG_SHORT_TYPE;
+	ival = (long)(short) lval;
+	break;
+      case INT_TOKEN:
+	stype = SLANG_INT_TYPE;
+	ival = (long)(int) lval;
+	break;
+      case LONG_TOKEN:
+	stype = LONG_TOKEN;
+        ival = lval;
+	break;
+
+      default:
+	return 0;
+     }
+
+   if (ival == lval)
+     {
+	if (((lval >= 0) && (sign > 0))
+	    || ((lval < 0) && (sign < 0)))
+	  return 0;
+     }
+   SLang_verror (SL_SYNTAX_ERROR, "Literal integer constant is too large for %s", SLclass_get_datatype_name(stype));
+   return -1;
+}
+
+static int check_uint_token_overflow (_pSLang_Token_Type *ctok, int sign)
+{
+   unsigned long ival, lval;
+   SLtype stype;
+
+   ctok->v.long_val = sign * ctok->v.long_val;
+   lval = (unsigned long) ctok->v.long_val;
+
+   switch (ctok->type)
+     {
+      case UCHAR_TOKEN:
+	ival = (unsigned long)(unsigned char) lval;
+	stype = SLANG_UCHAR_TYPE;
+	break;
+      case USHORT_TOKEN:
+	stype = SLANG_USHORT_TYPE;
+	ival = (unsigned long)(unsigned short) lval;
+	break;
+      case UINT_TOKEN:
+	stype = SLANG_UINT_TYPE;
+	ival = (unsigned long)(unsigned int) lval;
+	break;
+      case ULONG_TOKEN:
+	stype = SLANG_ULONG_TYPE;
+	ival = lval;
+	break;
+      default:
+	return 0;
+     }
+
+   if (ival == lval)
+     return 0;
+
+   SLang_verror (SL_SYNTAX_ERROR, "Literal integer constant is too large for %s", SLclass_get_datatype_name(stype));
+   return -1;
+}
+
+#ifdef HAVE_LONG_LONG
+static int check_llong_token_overflow (_pSLang_Token_Type *ctok, int sign)
+{
+   long long lval = sign * ctok->v.llong_val;
+   ctok->v.llong_val = lval;
+
+   if ((ctok->flags & SLTOKEN_IS_HEX)
+       || ((lval >= 0) && (sign > 0))
+       || ((lval < 0) && (sign < 0)))
+     return 0;
+
+   SLang_verror (SL_SYNTAX_ERROR, "Literal integer constant is too large for %s", SLclass_get_datatype_name(SLANG_LLONG_TYPE));
+   return -1;
+}
+
+static int check_ullong_token_overflow (_pSLang_Token_Type *ctok, int sign)
+{
+   ctok->v.ullong_val *= sign;
+   return 0;
+}
+#endif
+
+static int check_number_token_overflow (_pSLang_Token_Type *tok, int sign)
+{
+   tok->flags |= SLTOKEN_OVERFLOW_CHECKED;
+
+   switch (tok->type)
+     {
+      case CHAR_TOKEN:
+      case SHORT_TOKEN:
+      case INT_TOKEN:
+      case LONG_TOKEN:
+	return check_int_token_overflow (tok, sign);
+
+      case UCHAR_TOKEN:
+      case USHORT_TOKEN:
+      case UINT_TOKEN:
+      case ULONG_TOKEN:
+	return check_uint_token_overflow (tok, sign);
+
+#ifdef HAVE_LONG_LONG
+      case LLONG_TOKEN:
+	return check_llong_token_overflow (tok, sign);
+
+      case ULLONG_TOKEN:
+	return check_ullong_token_overflow (tok, sign);
+#endif
+     }
+   return 0;
+}
+
 static int compile_token (_pSLang_Token_Type *t)
 {
 #if SLANG_HAS_DEBUG_CODE
@@ -108,6 +239,11 @@ static int compile_token (_pSLang_Token_Type *t)
 	(*_pSLcompile_ptr) (&tok);
      }
 #endif
+   if ((t->flags & (SLTOKEN_TYPE_INTEGER|SLTOKEN_OVERFLOW_CHECKED)) == SLTOKEN_TYPE_INTEGER)
+     {
+	if (-1 == check_number_token_overflow (t, 1))
+	  return -1;
+     }
    (*_pSLcompile_ptr) (t);
    return 0;
 }
@@ -148,7 +284,7 @@ static void free_token_list (Token_List_Type *t)
 #if USE_PARANOID_MAGIC
    if (t->magic != 0xABCDEF12)
      {
-	SLang_verror (SL_INTERNAL_ERROR, "Token Magic number error.");
+	_pSLang_verror (SL_INTERNAL_ERROR, "Token Magic number error.");
 	return;
      }
 #endif
@@ -210,7 +346,7 @@ static int check_token_list_space (Token_List_Type *t, unsigned int delta_size)
 #if USE_PARANOID_MAGIC
    if (t->magic != 0xABCDEF12)
      {
-	SLang_verror (SL_INTERNAL_ERROR, "Token Magic number error.");
+	_pSLang_verror (SL_INTERNAL_ERROR, "Token Magic number error.");
 	return -1;
      }
 #endif
@@ -259,10 +395,12 @@ static int append_copy_of_string_token (_pSLang_Token_Type *t)
    t1 = Token_List->stack + Token_List->len;
    *t1 = *t;
    
-   if (NULL == (t1->v.s_val = SLang_create_slstring (t->v.s_val)))
+   if (t->v.s_val == NULL)
      return -1;
-   
-   t1->free_sval_flag = 1;
+
+   if (EOF_TOKEN == _pSLtoken_init_slstring_token (t1, t->type, t->v.s_val, strlen (t->v.s_val)))
+     return -1;
+
    t1->num_refs = 1;
 
    Token_List->len += 1;
@@ -275,6 +413,7 @@ static int append_int_token (int n)
 
    init_token (&num_tok);
    num_tok.type = INT_TOKEN;
+   num_tok.flags |= SLTOKEN_TYPE_INTEGER|SLTOKEN_OVERFLOW_CHECKED;
    num_tok.v.long_val = n;
    return append_token (&num_tok);
 }
@@ -289,6 +428,7 @@ static int append_token_of_type (unsigned char t)
    /* The memset when the list was created ensures that the other fields
     * are properly initialized.
     */
+#if 0
    if ((t == CHS_TOKEN) && Token_List->len)
      {
 	tok = Token_List->stack + (Token_List->len-1);
@@ -305,6 +445,7 @@ static int append_token_of_type (unsigned char t)
 	  }
 #endif
      }
+#endif
 
    tok = Token_List->stack + Token_List->len;
    init_token (tok);
@@ -442,6 +583,7 @@ static void compile_token_of_type (unsigned char t)
 #if SLANG_HAS_DEBUG_CODE
    tok.line_number = -1;
 #endif
+   tok.flags = 0;
    tok.type = t;
    compile_token(&tok);
 }
@@ -497,6 +639,7 @@ static int append_bos (_pSLang_Token_Type *t, int level)
 #endif
 
 static void statement (_pSLang_Token_Type *);
+static void loop_statement (_pSLang_Token_Type *);
 static void compound_statement (_pSLang_Token_Type *);
 static void expression_with_parenthesis (_pSLang_Token_Type *);
 static void handle_semicolon (_pSLang_Token_Type *);
@@ -516,6 +659,7 @@ static void postfix_expression (_pSLang_Token_Type *);
 static int check_for_lvalue (unsigned char, _pSLang_Token_Type *);
 /* static void primary_expression (_pSLang_Token_Type *); */
 static void block (_pSLang_Token_Type *);
+static void loop_block (_pSLang_Token_Type *);
 static void inline_array_expression (_pSLang_Token_Type *);
 static void inline_list_expression (_pSLang_Token_Type *);
 static void array_index_expression (_pSLang_Token_Type *);
@@ -559,7 +703,7 @@ static int get_identifier_token (_pSLang_Token_Type *tok)
 static void define_function (_pSLang_Token_Type *ctok, unsigned char type)
 {
    _pSLang_Token_Type fname;
-   
+
    switch (type)
      {
       case STATIC_TOKEN:
@@ -587,7 +731,12 @@ static void define_function (_pSLang_Token_Type *ctok, unsigned char type)
    compile_token_of_type(FARG_TOKEN);
 
    if (ctok->type == OBRACE_TOKEN)
-     compound_statement(ctok);
+     {
+	int loop_context = In_Looping_Context;
+	In_Looping_Context = 0;
+	compound_statement(ctok);
+	In_Looping_Context = loop_context;
+     }
 
    else if (ctok->type != SEMICOLON_TOKEN)
      {
@@ -604,15 +753,15 @@ static void define_function (_pSLang_Token_Type *ctok, unsigned char type)
 /* This is called from "statement", which is expected to return no
  * new token.
  */
-static void check_for_loop_then_else (_pSLang_Token_Type *ctok)
+static int check_for_loop_then_else (_pSLang_Token_Type *ctok)
 {
    int b = 0;
-   
+
    get_token (ctok);
 
    while (1)
      {
-	unsigned char type = ctok->type;
+	_pSLtok_Type type = ctok->type;
 #ifdef LOOP_ELSE_TOKEN
 	if ((type == ELSE_TOKEN) && ((b & 1) == 0))
 	  {
@@ -629,13 +778,18 @@ static void check_for_loop_then_else (_pSLang_Token_Type *ctok)
 	     get_token (ctok);
 	     block (ctok);
 	     compile_token_of_type (LOOP_THEN_TOKEN);
-	     get_token (ctok);
 	     b |= 2;
+#ifdef LOOP_ELSE_TOKEN
+	     get_token (ctok);
 	     continue;
+#else
+	     return b;
+#endif
 	  }
 	break;
      }
    unget_token (ctok);
+   return b;
 }
 
 /* statement:
@@ -725,7 +879,8 @@ static void statement (_pSLang_Token_Type *ctok)
       case _FOR_TOKEN:
 	get_token (ctok);
 	handle_for_statement (ctok);
-	check_for_loop_then_else (ctok);
+	if (0 == check_for_loop_then_else (ctok))
+	  compile_token_of_type (NOP_TOKEN);
 	break;
 
       case LOOP_TOKEN:
@@ -737,15 +892,17 @@ static void statement (_pSLang_Token_Type *ctok)
 #if SLANG_HAS_BOSEOS
 	if (eos) compile_eos ();
 #endif
-	block (ctok);
+	loop_block (ctok);
 	compile_token_of_type (LOOP_TOKEN);
-	check_for_loop_then_else (ctok);
+	if (0 == check_for_loop_then_else (ctok))
+	  compile_token_of_type (NOP_TOKEN);
 	break;
 
       case FOREACH_TOKEN:
 	get_token (ctok);
 	handle_foreach_statement (ctok);
-	check_for_loop_then_else (ctok);
+	if (0 == check_for_loop_then_else (ctok))
+	  compile_token_of_type (NOP_TOKEN);
 	break;
 
       case WHILE_TOKEN:
@@ -759,14 +916,15 @@ static void statement (_pSLang_Token_Type *ctok)
 	if (eos) compile_eos ();
 #endif
 	compile_token_of_type (CBRACE_TOKEN);
-	block (ctok);
+	loop_block (ctok);
 	compile_token_of_type (WHILE_TOKEN);
-	check_for_loop_then_else (ctok);
+	if (0 == check_for_loop_then_else (ctok))
+	  compile_token_of_type (NOP_TOKEN);
 	break;
 
       case DO_TOKEN:
 	get_token (ctok);
-	block (ctok);
+	loop_block (ctok);
 
 	if (WHILE_TOKEN != get_token (ctok))
 	  {
@@ -787,11 +945,11 @@ static void statement (_pSLang_Token_Type *ctok)
 	compile_token_of_type (CBRACE_TOKEN);
 	compile_token_of_type (DOWHILE_TOKEN);
 	handle_semicolon (ctok);
-	check_for_loop_then_else (ctok);
+	if (0 == check_for_loop_then_else (ctok))
+	  compile_token_of_type (NOP_TOKEN);
 	break;
 
       case FOR_TOKEN:
-
 	/* Look for (exp_opt ; exp_opt ; exp_opt ) */
 
 	if (OPAREN_TOKEN != get_token (ctok))
@@ -860,13 +1018,20 @@ static void statement (_pSLang_Token_Type *ctok)
 	compile_token_list ();
 
 	get_token (ctok);
-	block (ctok);
+	loop_block (ctok);
 	compile_token_of_type (FOR_TOKEN);
-	check_for_loop_then_else (ctok);
+	if (0 == check_for_loop_then_else (ctok))
+	  compile_token_of_type (NOP_TOKEN);
 	break;
 
       case FOREVER_TOKEN:
-	/* drop */
+	get_token (ctok);
+	loop_block (ctok);
+	compile_token_of_type (FOREVER_TOKEN);
+	if (0 == check_for_loop_then_else (ctok))
+	  compile_token_of_type (NOP_TOKEN);
+	break;
+
       case ERRBLK_TOKEN:
       case EXITBLK_TOKEN:
       case USRBLK0_TOKEN:
@@ -878,19 +1043,37 @@ static void statement (_pSLang_Token_Type *ctok)
 	get_token (ctok);
 	block (ctok);
 	compile_token_of_type (type);
-	if (type == FOREVER_TOKEN)
-	  check_for_loop_then_else (ctok);
 	break;
 
       case BREAK_TOKEN:
       case CONT_TOKEN:
+	if (In_Looping_Context == 0)
+	  {
+	     _pSLparse_error (SL_SYNTAX_ERROR, "`break' or `continue' requires a looping context", ctok, 0);
+	     break;
+	  }
+
+	type = ctok->type;
 #if SLANG_HAS_BOSEOS
 	eos = compile_bos (ctok, 3);
+#endif
+	if (SEMICOLON_TOKEN != get_token (ctok))
+	  {
+	     if ((ctok->type != INT_TOKEN)
+		 || (ctok->v.long_val <= 0))
+	       {
+		  _pSLparse_error (SL_SYNTAX_ERROR, "Expecting a positive non-zero integer or a semi-colon.", ctok, 0);
+		  break;
+	       }
+	     ctok->type = (type == BREAK_TOKEN) ? BREAK_N_TOKEN : CONT_N_TOKEN;
+	     compile_token (ctok);
+	     get_token (ctok);
+	     handle_semicolon (ctok);
+	  }
+	else compile_token_of_type (type);
+#if SLANG_HAS_BOSEOS
 	if (eos) compile_eos ();
 #endif
-	compile_token_of_type (ctok->type);
-	get_token (ctok);
-	handle_semicolon (ctok);
 	break;
 
       case RETURN_TOKEN:
@@ -1052,6 +1235,13 @@ static void statement (_pSLang_Token_Type *ctok)
    LLT->parse_level -= 1;
 }
 
+static void loop_statement (_pSLang_Token_Type *ctok)
+{
+   In_Looping_Context++;
+   statement (ctok);
+   In_Looping_Context--;
+}
+
 /* This function does not return a new token.  Calling routine must do that */
 static void block (_pSLang_Token_Type *ctok)
 {
@@ -1060,6 +1250,13 @@ static void block (_pSLang_Token_Type *ctok)
    compile_token_of_type (CBRACE_TOKEN);
 }
 
+static void loop_block (_pSLang_Token_Type *ctok)
+{
+   compile_token_of_type (OBRACE_TOKEN);
+   loop_statement (ctok);
+   compile_token_of_type (CBRACE_TOKEN);
+}
+
 static void handle_for_statement (_pSLang_Token_Type *ctok)
 {
    _pSLang_Token_Type tok_buf;
@@ -1089,32 +1286,19 @@ static void handle_for_statement (_pSLang_Token_Type *ctok)
 	compile_token (ident_token);
 	free_token (ident_token);
      }
-   statement (ctok);
+   loop_statement (ctok);
    compile_token_of_type (CBRACE_TOKEN);
 
    compile_token_of_type (_FOR_TOKEN);
 }
 
-static int init_identifier_token (_pSLang_Token_Type *t, char *name, int alloc)
+static int init_identifier_token (_pSLang_Token_Type *t, SLFUTURE_CONST char *name)
 {
-   unsigned int len;
-   
-   len = strlen (name);
-
    init_token (t);
-   if (alloc)
-     {
-	if (NULL == (t->v.s_val = _pSLstring_make_hashed_string (name, len, &t->hash)))
-	  return -1;
+
+   if (EOF_TOKEN == _pSLtoken_init_slstring_token (t, IDENT_TOKEN, name, strlen(name)))
+     return -1;
    
-	t->free_sval_flag = 1;
-	return 0;
-     }
-   
-   t->type = IDENT_TOKEN;
-   t->v.s_val = name;
-   t->hash = _pSLstring_hash ((unsigned char *)name, (unsigned char *)name + len);
-   t->free_sval_flag = 0;
    return 0;
 }
 
@@ -1160,7 +1344,7 @@ static void handle_try_statement (_pSLang_Token_Type *ctok)
    if (ctok->type == OPAREN_TOKEN)
      {
 	_pSLang_Token_Type e;
-	if (-1 == init_identifier_token (&e, "__get_exception_info", 0))
+	if (-1 == init_identifier_token (&e, "__get_exception_info"))
 	  return;
 	
 	append_token (&e);
@@ -1364,7 +1548,9 @@ static void handle_foreach_statement (_pSLang_Token_Type *ctok)
 	compile_token (v);
 	v = v->next;
      }
-   statement (ctok);
+
+   loop_statement (ctok);
+
    compile_token_of_type (CBRACE_TOKEN);
    compile_token_of_type (FOREACH_TOKEN);
    
@@ -1448,6 +1634,7 @@ void _pSLparse_start (SLang_Load_Type *llt)
    unsigned int save_use_next_token;
    _pSLang_Token_Type save_next_token;
    Token_List_Type *save_list;
+   int save_looping_context = In_Looping_Context;
 #if SLANG_HAS_DEBUG_CODE
    int save_last_line_number = Last_Line_Number;
 
@@ -1461,15 +1648,20 @@ void _pSLparse_start (SLang_Load_Type *llt)
 
    init_token (&Next_Token);
    Use_Next_Token = 0;
+   In_Looping_Context = 0;
    init_token (&ctok);
    get_token (&ctok);
 
    llt->parse_level = 0;
    statement_list (&ctok);
 
-   if ((_pSLang_Error == 0)
-       && (ctok.type != EOF_TOKEN))
-     _pSLparse_error (SL_SYNTAX_ERROR, "Parse ended prematurely", &ctok, 0);
+   if (_pSLang_Error == 0)
+     {
+	if (ctok.type != EOF_TOKEN)
+	  _pSLparse_error (SL_SYNTAX_ERROR, "Parse ended prematurely", &ctok, 0);
+	else
+	  compile_token_of_type (EOF_TOKEN);
+     }
    
 
    if (_pSLang_Error)
@@ -1490,6 +1682,8 @@ void _pSLparse_start (SLang_Load_Type *llt)
      free_token (&Next_Token);
    Use_Next_Token = save_use_next_token;
    Next_Token = save_next_token;
+   In_Looping_Context = save_looping_context;
+
 #if SLANG_HAS_DEBUG_CODE
    Last_Line_Number = save_last_line_number;
 #endif
@@ -1566,7 +1760,7 @@ static void free_token_linked_list (_pSLang_Token_Type *tok)
 	free_token (tok);
 	if (tok->num_refs != 0)
 	  {
-	     SLang_verror (SL_INTERNAL_ERROR, "Cannot free token in linked list");
+	     _pSLang_verror (SL_INTERNAL_ERROR, "Cannot free token in linked list");
 	  }
 	else
 	  SLfree ((char *) tok);
@@ -1653,7 +1847,7 @@ static _pSLang_Token_Type *
    
    if (n == 0)
      {
-	_pSLparse_error (SL_SYNTAX_ERROR, "Expecting an identifier", ctok, 0);
+	_pSLparse_error (SL_SYNTAX_ERROR, "Expecting a qualifier", ctok, 0);
 	return NULL;
      }
 
@@ -2082,6 +2276,7 @@ static void handle_binary_sequence (_pSLang_Token_Type *ctok, unsigned char max_
 /* % Note: simple-expression groups operators OP1 at same level.  The
  * % actual implementation will not do this.
  * simple-expression:
+ *       simple-expression ? simple-expression : simple-expression
  *	 unary-expression
  *	 binary-expression BINARY-OP unary-expression
  *       andelse xxelse-expression-list
@@ -2140,6 +2335,45 @@ static void simple_expression (_pSLang_Token_Type *ctok)
      return;
 
    handle_binary_sequence (ctok, 0xFF);
+   
+   if (ctok->type == QUESTION_TOKEN)
+     {
+	append_token_of_type (OBRACE_TOKEN);
+	get_token (ctok);
+	simple_expression (ctok);
+	if (ctok->type != COLON_TOKEN)
+	  {
+	     _pSLparse_error (SL_SYNTAX_ERROR, "Expecting a colon in the ternary expression", ctok, 0);
+	     return;
+	  }
+	append_token_of_type (CBRACE_TOKEN);
+	get_token (ctok);
+	append_token_of_type (OBRACE_TOKEN);
+	simple_expression (ctok);
+	append_token_of_type (CBRACE_TOKEN);
+	append_token_of_type (ELSE_TOKEN);
+     }
+}
+
+
+static int negate_float_type_token (_pSLang_Token_Type *tok)
+{
+   char buf[SL_MAX_TOKEN_LEN];
+   unsigned int len;
+
+   len = 1 + _pSLstring_bytelen (tok->v.s_val);   /* sign */
+   if (len >= sizeof(buf))
+     {
+	_pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "Number too long for buffer", tok, 1);
+	return -1;
+     }
+   buf[0] = '-';
+   memcpy (buf+1, tok->v.s_val, len);  /* copys \0 */
+   (*tok->free_val_func)(tok);
+   if (EOF_TOKEN == _pSLtoken_init_slstring_token (tok, tok->type, buf, len))
+     return -1;
+
+   return 0;
 }
 
 /* unary-expression:
@@ -2179,20 +2413,29 @@ static void unary_expression (_pSLang_Token_Type *ctok)
 
 	   case SUB_TOKEN:
 	     (void) get_token (ctok);
-#if 0
-	     if (IS_INTEGER_TOKEN (ctok->type))
+	     if (ctok->flags & SLTOKEN_TYPE_NUMBER)
 	       {
-		  ctok->v.long_val = -ctok->v.long_val;
-		  break;
+		  _pSLang_Token_Type *last_token;
+		  postfix_expression (ctok);
+		  if ((NULL != (last_token = get_last_token ()))
+		      && (last_token->flags & SLTOKEN_TYPE_NUMBER))
+		    {
+		       if (last_token->flags & SLTOKEN_TYPE_FLOAT)
+			 {
+			    if (-1 == negate_float_type_token (last_token))
+			      return;
+			 }
+		       else if (-1 == check_number_token_overflow (last_token, -1))
+			 return;
+		    }
+		  else
+		    {
+		       if (num_unary_ops == 16)
+			 goto stack_overflow_error;
+		       save_unary_ops [num_unary_ops++] = CHS_TOKEN;
+		    }
+		  goto out_of_switch;
 	       }
-#ifdef HAVE_LONG_LONG
-	     else if ((ctok->type == LLONG_TOKEN) || (ctok->type == ULLONG_TOKEN))
-	       {
-		  ctok->v.llong_val = -ctok->v.llong_val;
-		  break;
-	       }
-#endif
-#endif
 	     if (num_unary_ops == 16)
 	       goto stack_overflow_error;
 	     save_unary_ops [num_unary_ops++] = CHS_TOKEN;
@@ -2238,7 +2481,8 @@ static void unary_expression (_pSLang_Token_Type *ctok)
 
 static int combine_namespace_tokens (_pSLang_Token_Type *a, _pSLang_Token_Type *b)
 {
-   char *sa, *sb, *sc;
+   SLFUTURE_CONST char *sa, *sb;
+   char *sc;
    unsigned int lena, lenb;
    unsigned long hash;
 
@@ -2270,7 +2514,7 @@ static int combine_namespace_tokens (_pSLang_Token_Type *a, _pSLang_Token_Type *
    /* I can free this string because no other token should be referencing it.
     * (num_refs == 1).
     */
-   _pSLfree_hashed_string (sa, lena, a->hash);
+   _pSLfree_hashed_string ((char *) sa, lena, a->hash);
    a->v.s_val = sb;
    a->hash = hash;
 
@@ -2292,7 +2536,12 @@ static void append_identifier_token (_pSLang_Token_Type *ctok)
 	return;
      }
 
-   last_token = get_last_token ();
+   if (NULL == (last_token = get_last_token ()))
+     {
+	if (_pSLang_Error == 0)
+	  _pSLang_verror (SL_INTERNAL_ERROR, "get_last_token returned NULL in append_identifier_token");
+	return;
+     }
    if (-1 == combine_namespace_tokens (last_token, ctok))
      return;
 
@@ -2389,6 +2638,10 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
       case USHORT_TOKEN:
       case UINT_TOKEN:
       case ULONG_TOKEN:
+#ifdef HAVE_LONG_LONG
+      case LLONG_TOKEN:
+      case ULLONG_TOKEN:
+#endif
       case STRING_TOKEN:
       case BSTRING_TOKEN:
 #ifdef SLANG_HAS_FLOAT
@@ -2400,10 +2653,7 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
       case COMPLEX_TOKEN:
 #endif
       case STRING_DOLLAR_TOKEN:
-#ifdef HAVE_LONG_LONG
-      case LLONG_TOKEN:
-      case ULLONG_TOKEN:
-#endif
+      case MULTI_STRING_TOKEN:
 	append_token (ctok);
 	get_token (ctok);
 	break;
@@ -2432,6 +2682,12 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
 	get_token (ctok);
 	postfix_expression (ctok);
 	last_token = get_last_token ();
+	if (last_token == NULL)
+	  {
+	     if (_pSLang_Error == 0)
+	       _pSLang_verror (SL_SYNTAX_ERROR, "Misplaced &");
+	     return;
+	  }
 	switch (last_token->type)
 	  {
 	   case IDENT_TOKEN:
