@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004, 2005 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -60,9 +60,19 @@ SLSig_Fun_Type *SLsignal (int sig, SLSig_Fun_Type *f)
 # ifdef SA_RESTART
    new_sa.sa_flags |= SA_RESTART;
 # endif
-
-   if (-1 == sigaction (sig, &new_sa, &old_sa))
-     return (SLSig_Fun_Type *) SIG_ERR;
+   
+   while (-1 == sigaction (sig, &new_sa, &old_sa))
+     {
+#ifdef EINTR
+	if  (errno == EINTR)
+	  {
+	     if (0 == SLang_handle_interrupt ())
+	       continue;
+	  }
+#endif
+	_pSLerrno_errno = errno;
+	return (SLSig_Fun_Type *) SIG_ERR;
+     }
 
    return old_sa.sa_handler;
 #else
@@ -87,8 +97,18 @@ SLSig_Fun_Type *SLsignal_intr (int sig, SLSig_Fun_Type *f)
    new_sa.sa_flags |= SA_INTERRUPT;
 # endif
 
-   if (-1 == sigaction (sig, &new_sa, &old_sa))
-     return (SLSig_Fun_Type *) SIG_ERR;
+   while (-1 == sigaction (sig, &new_sa, &old_sa))
+     {
+# ifdef EINTR
+	if  (errno == EINTR)
+	  {
+	     if (0 == SLang_handle_interrupt ())
+	       continue;
+	  }
+#endif
+	_pSLerrno_errno = errno;
+	return (SLSig_Fun_Type *) SIG_ERR;
+     }
 
    return old_sa.sa_handler;
 #else
@@ -167,14 +187,18 @@ int SLsig_unblock_signals (void)
 }
 
 #ifdef MSWINDOWS
-int SLsystem (char *cmd)
+int SLsystem (SLFUTURE_CONST char *cmd)
 {
-   SLang_verror (SL_NOT_IMPLEMENTED, "system not implemented");
+   _pSLang_verror (SL_NOT_IMPLEMENTED, "system not implemented");
    return -1;
 }
+int SLsystem_intr (SLFUTURE_CONST char *cmd)
+{
+   return SLsystem (cmd);
+}
 
 #else
-int SLsystem (char *cmd)
+static int system_internal (SLFUTURE_CONST char *cmd, int do_sigint)
 {
 #ifdef SLANG_POSIX_SIGNALS
    pid_t pid;
@@ -197,14 +221,16 @@ int SLsystem (char *cmd)
    ignore.sa_flags = 0;
 
 # ifdef SIGINT
-   if (-1 == sigaction (SIGINT, &ignore, &save_intr))
+   if (do_sigint
+       && (-1 == sigaction (SIGINT, &ignore, &save_intr)))
      return -1;
 # endif
 
 # ifdef SIGQUIT
    if (-1 == sigaction (SIGQUIT, &ignore, &save_quit))
      {
-	(void) sigaction (SIGINT, &save_intr, NULL);
+	if (do_sigint)
+	  (void) sigaction (SIGINT, &save_intr, NULL);
 	return -1;
      }
 # endif
@@ -215,7 +241,8 @@ int SLsystem (char *cmd)
    if (-1 == sigprocmask (SIG_BLOCK, &child_mask, &save_mask))
      {
 #  ifdef SIGINT
-	(void) sigaction (SIGINT, &save_intr, NULL);
+	if (do_sigint)
+	  (void) sigaction (SIGINT, &save_intr, NULL);
 #  endif
 #  ifdef SIGQUIT
 	(void) sigaction (SIGQUIT, &save_quit, NULL);
@@ -227,12 +254,16 @@ int SLsystem (char *cmd)
    pid = fork();
 
    if (pid == -1)
-     status = -1;
+     {
+	_pSLerrno_errno = errno;
+	status = -1;
+     }
    else if (pid == 0)
      {
 	/* Child */
 # ifdef SIGINT
-	(void) sigaction (SIGINT, &save_intr, NULL);
+	if (do_sigint)
+	  (void) sigaction (SIGINT, &save_intr, NULL);
 # endif
 # ifdef SIGQUIT
 	(void) sigaction (SIGQUIT, &save_quit, NULL);
@@ -251,18 +282,26 @@ int SLsystem (char *cmd)
 	  {
 # ifdef EINTR
 	     if (errno == EINTR)
-	       continue;
+	       {
+		  (void) SLang_handle_interrupt ();
+		  continue;
+	       }
 # endif
 # ifdef ERESTARTSYS
 	     if (errno == ERESTARTSYS)
-	       continue;
+	       {
+		  (void) SLang_handle_interrupt ();
+		  continue;
+	       }
 # endif
+	     _pSLerrno_errno = errno;
 	     status = -1;
 	     break;
 	  }
      }
 # ifdef SIGINT
-   if (-1 == sigaction (SIGINT, &save_intr, NULL))
+   if (do_sigint
+       && (-1 == sigaction (SIGINT, &save_intr, NULL)))
      status = -1;
 # endif
 # ifdef SIGQUIT
@@ -289,11 +328,16 @@ int SLsystem (char *cmd)
    squit = SLsignal (SIGQUIT, SIG_IGN);
 # endif
 # ifdef SIGINT
-   sint = SLsignal (SIGINT, SIG_IGN);
+   if (do_sigint)
+     sint = SLsignal (SIGINT, SIG_IGN);
 # endif
+
    status = system (cmd);
+   _pSLerrno_errno = errno;
+
 # ifdef SIGINT
-   SLsignal (SIGINT, sint);
+   if (do_sigint)
+     SLsignal (SIGINT, sint);
 # endif
 # ifdef SIGQUIT
    SLsignal (SIGQUIT, squit);
@@ -301,8 +345,20 @@ int SLsystem (char *cmd)
    return status;
 #endif				       /* POSIX_SIGNALS */
 }
+
+int SLsystem (SLFUTURE_CONST char *cmd)
+{
+   return system_internal (cmd, 1);
+}
+
+int SLsystem_intr (SLFUTURE_CONST char *cmd)
+{
+   return system_internal (cmd, 0);
+}
+
 #endif
 
+
 #if 0
 #include <windows.h>
 static int msw_system (char *cmd)
@@ -329,7 +385,7 @@ static int msw_system (char *cmd)
 			       &startup_info,
 			       &process_info))
      {
-	SLang_verror (0, "%s: CreateProcess failed.", cmd);
+	_pSLang_verror (0, "%s: CreateProcess failed.", cmd);
 	return -1;
      }
 
