@@ -1,3 +1,24 @@
+/* slwclut.c: wide character lookup tables */
+/*
+Copyright (C) 2004-2009 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 #include "slinclud.h"
 #include <string.h>
 
@@ -15,7 +36,7 @@ struct SLwchar_Lut_Type
    SLwchar_Type *chmin, *chmax;
    unsigned int table_len;
    unsigned int malloced_len;
-   unsigned char char_class;
+   unsigned int char_class;
 };
 
 void SLwchar_free_lut (SLwchar_Lut_Type *r)
@@ -104,7 +125,7 @@ int SLwchar_add_range_to_lut (SLwchar_Lut_Type *r, SLwchar_Type a, SLwchar_Type
    return 0;
 }
 
-static void add_char_class (SLwchar_Lut_Type *r, unsigned char char_class)
+static void add_char_class (SLwchar_Lut_Type *r, unsigned int char_class)
 {
    unsigned int i;
    unsigned char *lut;
@@ -292,7 +313,7 @@ SLuchar_Type *SLwchar_bskip_range (SLwchar_Lut_Type *r, SLuchar_Type *pmin,
 
 typedef struct
 {
-   char *name;
+   SLCONST char *name;
    char escaped_form;
 }
 Posix_Char_Class_Type;
@@ -327,7 +348,7 @@ static int is_posix_charclass (SLuchar_Type **up, SLuchar_Type *umax, SLwchar_Ty
 
    u1 = u;
    while ((u1 < umax)
-	  && (*u1 > 'a')
+	  && (*u1 >= 'a')
 	  && (*u1 <= 'z'))
      u1++;
 
@@ -347,7 +368,7 @@ static int is_posix_charclass (SLuchar_Type **up, SLuchar_Type *umax, SLwchar_Ty
 	  }
 	p++;
      }
-   SLang_verror (SL_NotImplemented_Error, "Character class in range specification is unknown or unsupported");
+   _pSLang_verror (SL_NotImplemented_Error, "Character class in range specification is unknown or unsupported");
    return -1;
 }
 
@@ -391,7 +412,15 @@ static int get_lex_char (SLuchar_Type **up, SLuchar_Type *umax,
 	*up = u;
 	return 0;
      }
-   if (*u == '\\')
+   
+   /* Here, ch=='\\' and *u represents the next character. */
+
+   /* Allow \\ and \^ to represent \ and ^, resp.  Supporting \^ is useful 
+    * in constructs such as "\\^x" since "^x" may mean anything but x, and not
+    * '^' or 'x'.
+    */
+   ch = *u;
+   if ((ch == '\\') || (ch == '^'))
      {
 	*char_classp = 0;
 	*chp = ch;
@@ -442,6 +471,10 @@ static SLuchar_Type *get_lexical_element (SLuchar_Type *u, SLuchar_Type *umax,
 	lex->lexical_type = LEXICAL_CLASS_TYPE;
 	switch (char_class)
 	  {
+	   case '7':
+	     lex->e.char_class = SLCHARCLASS_ASCII;
+	     break;
+
 	   case 'a':	       /* alpha */
 	     lex->e.char_class = SLCHARCLASS_ALPHA;
 	     break;
@@ -491,6 +524,10 @@ static SLuchar_Type *get_lexical_element (SLuchar_Type *u, SLuchar_Type *umax,
 	   case 'w':	       /* alphanumeric */
 	     lex->e.char_class = SLCHARCLASS_ALPHA|SLCHARCLASS_XDIGIT;
 	     break;
+	     
+	   default:
+	     _pSLang_verror (SL_INVALID_PARM, "Invalid character class '%c'.", char_class);
+	     return NULL;
 	  }
 	return u;
      }
@@ -508,7 +545,7 @@ static SLuchar_Type *get_lexical_element (SLuchar_Type *u, SLuchar_Type *umax,
 	lex->lexical_type = LEXICAL_CHAR_TYPE;
 	lex->e.wch = '-';
 	return u;
-	/* SLang_verror (SL_INVALID_PARM, "Unfinished range specification"); */
+	/* _pSLang_verror (SL_INVALID_PARM, "Unfinished range specification"); */
 	/* return NULL; */
      }
 
@@ -517,13 +554,13 @@ static SLuchar_Type *get_lexical_element (SLuchar_Type *u, SLuchar_Type *umax,
    
    if (char_class)
      {
-	SLang_verror (SL_INVALID_PARM, "Character class not allowed in a range");
+	_pSLang_verror (SL_INVALID_PARM, "Character class not allowed in a range");
 	return NULL;
      }
 
    if (r1 == 0)
      {
-	SLang_verror (SL_INVALID_PARM, "Unfinished range specification");
+	_pSLang_verror (SL_INVALID_PARM, "Unfinished range specification");
 	return NULL;
      }
    
@@ -689,6 +726,9 @@ static int is_of_class (int char_class, SLwchar_Type w)
 	
       case SLCHARCLASS_SPACE:
 	return SLwchar_isspace (w);
+	
+      case SLCHARCLASS_ASCII:
+	return w < (SLwchar_Type)0x80;
      }
    
    return 0;
@@ -764,6 +804,11 @@ static void get_range_values (Lexical_Element_Type *lex,
    *chmaxp = chmax;
 }
 
+static int is_ascii (SLwchar_Type wch)
+{
+   return wch < (SLwchar_Type) 0x80;
+}
+
 static int check_char_mapping (SLwchar_Map_Type *map, Char_Map_Type *list, int first_time)
 {
    Lexical_Element_Type *lex_from, *lex_to;
@@ -844,7 +889,7 @@ static int check_char_mapping (SLwchar_Map_Type *map, Char_Map_Type *list, int f
 	   case LEXICAL_RANGE_TYPE:
 	     if (invert)
 	       {
-		  SLang_verror (SL_INVALID_PARM, "Inversion from a range to a range not permitted");
+		  _pSLang_verror (SL_INVALID_PARM, "Inversion from a range to a range not permitted");
 		  return -1;
 	       }
 
@@ -852,7 +897,7 @@ static int check_char_mapping (SLwchar_Map_Type *map, Char_Map_Type *list, int f
 
 	     if ((chmax - chmin) != (wch1 - wch))
 	       {
-		  SLang_verror (SL_INVALID_PARM, "Character mapping of unequal ranges is forbidden");
+		  _pSLang_verror (SL_INVALID_PARM, "Character mapping of unequal ranges is forbidden");
 		  return -1;
 	       }
 	     if (from_range_dir != to_range_dir)
@@ -901,8 +946,12 @@ static int check_char_mapping (SLwchar_Map_Type *map, Char_Map_Type *list, int f
 	     is_func = SLwchar_isspace;
 	     break;
 		  
+	   case SLCHARCLASS_ASCII:
+	     is_func = is_ascii;
+	     break;
+
 	   default:
-	     SLang_verror (SL_INVALID_PARM, "Invalid character class in character map");
+	     _pSLang_verror (SL_INVALID_PARM, "Invalid character class in character map");
 	     return -1;
 	  }
 	switch (lex_to->lexical_type)
@@ -993,7 +1042,7 @@ SLwchar_Map_Type *SLwchar_allocate_char_map (SLuchar_Type *from, SLuchar_Type *t
 #if 0
    if (*from == 0)
      {
-	SLang_verror (SL_INVALID_PARM, "Illegal empty string in character map specification");
+	_pSLang_verror (SL_INVALID_PARM, "Illegal empty string in character map specification");
 	return NULL;
      }
 #endif
@@ -1039,7 +1088,7 @@ SLwchar_Map_Type *SLwchar_allocate_char_map (SLuchar_Type *from, SLuchar_Type *t
 	  {
 	     if (invert)
 	       {
-		  SLang_verror (SL_INVALID_PARM, "Character map inversion must specify a many-to-one mapping");
+		  _pSLang_verror (SL_INVALID_PARM, "Character map inversion must specify a many-to-one mapping");
 		  goto return_error;
 	       }
 	     to = next_to;
@@ -1055,7 +1104,7 @@ SLwchar_Map_Type *SLwchar_allocate_char_map (SLuchar_Type *from, SLuchar_Type *t
 
 	if (-1 == check_char_mapping (map, list, first_time))
 	  {
-	     SLang_verror (SL_INVALID_PARM, "Specified character mapping is invalid");
+	     _pSLang_verror (SL_INVALID_PARM, "Specified character mapping is invalid");
 	     goto return_error;
 	  }
 	first_time = 0;
